@@ -9,44 +9,12 @@
  *
  * I reused the scube code for lighting and colors.
  *
- * (c) Copyright 1995, Michael Hamilton
+ * (c) Copyright 1999, Michael Hamilton, Gentoo South Pacific Limited
  *
  * You have the same rights as stated in the original SGI copyright.
  * (see below).
  *
- * Input data format:  A continous series of data sets each of which has
- * the following format (newlines are significant).
- *
- *   title
- *   nitems nvalues valuename1 valuename2 valuename3 ...
- *   itemID
- *   groupID
- *   value1 value2 value3 ...
- *   repeat of the above ...
- *
- * E.g.
- *
- *   Selected processes on sputnik3
- *   4 3 CPU RSS MEM
- *   100
- *   michael
- *   0.000000 0.043189 0.000000 
- *   196
- *   root
- *   0.166667 0.559287 0.178711 
- *   250
- *   michael
- *   2.083333 1.768806 0.609375 
- *   275
- *   michael
- *   0.000000 0.054172 0.021484 
- *
- *   Note titles, nitems, nvalues can change for each data set.
- *   Data within a data set can be in any order, the program will
- *   sort it by ID and group ID (in the previous example ID is process ID,
- *   and group ID is process username).
- *
- *   Accepts data from stdin (e.g. rsh output).
+ * See gr_monitor.1 man page for usage.
  */
 
 /* ---- ORIGINAL scube.c stuff: */
@@ -96,7 +64,7 @@
 
 /* ---- End of original scube.c stuff. */
 
-#define VERSION_STRING "Gr_Monitor Version 0.53"
+#define VERSION_STRING "Gr_Monitor Version 0.80"
 
 #include <sys/types.h>
 #include <stdio.h>
@@ -128,12 +96,15 @@
 #include <GL/gl.h>
 #include <GL/glut.h>
 
+#include "dataset.h"
+
 #define DEBUG_MIN 1
 #define DEBUG_MAX 3
 
 #define ROTATION      6
-#define PER_ROW      18
-#define START_ANGLE  55
+#define PER_ROW      12
+#define START_ANGLE  50
+#define START_X      0
 #define START_Y      -3
 #define START_DOLLY  -9
 #define FONT_SCALE   0.002
@@ -142,9 +113,9 @@
 
 #define GR_GATHER_COMMAND "GR_GATHER_COMMAND"
 
-#ifndef GATHER_COMMAND
-#define GATHER_COMMAND "/usr/local/bin/gr_gather"
-#endif
+#define GATHER_PROC BINDIR "/gr_gather"
+#define GATHER_FREE BINDIR "/gr_free"
+#define GATHER_RUP  BINDIR "/gr_rup"
 
 #define MAX_LINE        10000
 #define MAX_VALUES        200
@@ -198,7 +169,7 @@ static int halted = 0;
 static int skip = 1;
 
 static int cram = 0;
-static int per_row = PER_ROW;
+static int per_row = 0;
 static int gap = ROW_GAP;
 static float cram_factor = CRAM_FACTOR;
 
@@ -209,21 +180,19 @@ static int rotate_z = 0;
 static int rotate_angle = START_ANGLE;
 
 static int trans_y = START_Y;
-static int trans_x = 0;
+static int trans_x = START_X;
 static int trans_z = START_DOLLY;
 
 static float zoom_horz = 1.0;
 static float zoom_vert = 1.0;
 
-static char value_name[MAX_VALUES][MAX_TEXT + 1];
-
 static int hide_value[MAX_CONTROLLABLE] = {
   0,0,0,0,0, 0,0,0,0,0 
 };
 
 static float scale_value[MAX_CONTROLLABLE] = {
-  1.0, 1.0, 1.0, 1.0, 1.0, 
-  1.0, 1.0, 1.0, 1.0, 1.0 
+  0.8, 0.8, 0.8, 0.8, 0.8, 
+  0.8, 0.8, 0.8, 0.8, 0.8 
 };
 
 static int value_index = 0;
@@ -334,58 +303,10 @@ static float cube_normals[6][4] =
 };
 
 
-typedef struct {
-  char id[MAX_TEXT + 1];
-  char group[MAX_TEXT + 1];
-  int group_size;
-  float *value;
-} item_t;
-
-
 void projection(float, float);
-void display(item_t *, int, int);
+void display(Dataset *);
 
-item_t *allocate_data (int num_items,
-		       int num_values)
-{
-  static item_t *allocated_store = NULL;
-  static int allocated_items = 0;
-  static int allocated_values = 0;
-
-  int i;
-
-  if (num_items != allocated_items || num_values != allocated_values) {
-    if (allocated_store) {
-      for (i=0; i < allocated_items; i++) {
-	free(allocated_store[i].value);
-      }
-      free((void *)allocated_store);
-    }
-    allocated_store = malloc (num_items * sizeof (item_t));
-    assert(allocated_store != NULL);
-    for (i=0; i < num_items; i++) {
-      allocated_store[i].value = malloc (sizeof(float) * num_values);
-    }
-    allocated_items = num_items;
-    allocated_values = num_values;
-  }
-  
-  return allocated_store;
-}
-
-
-int compareData(const void *arg1, const void *arg2)
-{
-  register const item_t *item1 = (item_t *) arg1;
-  register const item_t *item2 = (item_t *) arg2;
-  register int result;
-  if ((result = strcmp(item1->group, item2->group)) != 0) {
-    return result;
-  }
-  return strcmp(item1->id, item2->id);
-}
-
-int DataReady(FILE *inputfp) 
+int is_data_ready(FILE *inputfp) 
 {
   /* For select(2). */
   struct timeval tv;
@@ -397,18 +318,14 @@ int DataReady(FILE *inputfp)
   tv.tv_usec = 0;
   FD_ZERO(&in);
   FD_SET(fileno(inputfp), &in);
-  return select(16, &in, 0, 0, &tv) > 0;
+  return select(16, &in, 0, 0, &tv) > 0 && !feof(inputfp);
 }
 
-item_t *getData(FILE *inputfp,
-		char *title,
-		int *num_items, 
-		int *num_values, 
-		int prev_values) {
+Dataset *get_data(FILE *inputfp,
+		  int prev_values) {
 
-  static item_t *data_set = NULL;
-
-  static int prev_num_items, prev_num_values;
+  static Dataset *data_set = NULL;
+  Dataset *new_set = NULL;
 
   int i, j;
   int id, group_size;
@@ -417,104 +334,76 @@ item_t *getData(FILE *inputfp,
  
   static char line[MAX_LINE], *upto;
 
-  if ((prev_values || !DataReady(inputfp)) && data_set) {
+  if ((prev_values || !is_data_ready(inputfp)) && data_set) {
     if (debug >= DEBUG_MAX) fputs("Returning old data.\n", stderr);
-    *num_items  = prev_num_items;
-    *num_values = prev_num_values;
     return data_set;
   }
 
   /* Data is ready or we have no existing data-set */
 
-  if (debug >= DEBUG_MAX) fputs("Getting new data.\n", stderr);
-
-  fgets(title, MAX_TEXT, inputfp);
-
-  fgets(line, MAX_LINE, inputfp);
-  
-  *num_items = strtol(line, &upto, 0);
-  *num_values = strtol(upto, &upto, 0);
-
-  assert (*num_values <= MAX_VALUES);
-
-  for (i = 0; i < *num_values; i++) {
-    strncpy(value_name[i],strtok(upto," "),MAX_TEXT);
-    value_name[i][MAX_TEXT] = '\0';
-    upto = NULL;
-  }
-
-  data_set = allocate_data (*num_items, *num_values);
-  
-  for (i = 0; i < *num_items; i++) { 
-
-    fgets(data_set[i].id, MAX_TEXT, inputfp);
-    fgets(data_set[i].group, MAX_TEXT, inputfp);
-
-    fgets(line,MAX_LINE,inputfp);
-        
-    for (j = 0, upto=line; j < *num_values; j++) {
-      data_set[i].value[j] = strtod(upto, &upto); 
+  new_set = getDataset(inputfp);
+  if (new_set != NULL) {
+    if (data_set != NULL) {
+      freeDataset(data_set);
     }
-  }
-
-  qsort(data_set, *num_items, sizeof(item_t), &compareData) ;
-
-  strcpy(group, data_set[(*num_items) - 1].group) ;
-  group_size = 1;
-  for (i = (*num_items) - 1; i >= 0; i--) { 
-    if (strcmp(data_set[i].group, group) != 0) {
-      strcpy(group,data_set[i].group);
-      group_size = 1;
-    }
-    /* printf("gs=%d\n",group_size); */
-    data_set[i].group_size = group_size;
-    group_size++;
+    data_set = new_set;
   }
 
   if (cram) {
     
   }
 
-  if (debug >= DEBUG_MAX) {
-    fprintf(stderr, "title=%s num_items=%d num_values=%d\n", title, *num_items, *num_values);
-  }
-
-  prev_num_items  = *num_items;
-  prev_num_values = *num_values;
- 
   return data_set;
 }
 
 void do_update(void)
 {
+  static int first_time_thru = 1;
   static int count = 0;
   static char title[MAX_TEXT];
 
   int num_items;
   int num_values;
 
-  item_t *data_set = getData(inputfp, title, &num_items, &num_values, halted);
+  Dataset *data_set = get_data(inputfp, halted);
+
+  if (first_time_thru) {
+    first_time_thru = 0;
+    if (per_row == 0) {
+      if (data_set->groupCount > 0) {
+	per_row = data_set->groups[0]->itemCount;
+	/*printf("%d\n", per_row);*/
+      }
+      if (per_row < PER_ROW) {
+	per_row = PER_ROW;
+      }
+    }
+  }
+
 
   ++count;
 
-  if (skip) {
-    while (DataReady(inputfp)) { /* Skip over data until we are up to date */
+  if (skip && !halted) {
+    while (is_data_ready(inputfp)) { /* Skip over data until we are up to date */
       ++count;
       if (debug >= DEBUG_MIN) {
 	fprintf(stderr,
 		"Can't render fast enough - skipping over data set %d.\n",
 		count);
       }
-      data_set = getData(inputfp, title, &num_items, &num_values, halted);
+      data_set = get_data(inputfp, halted);
     } 
   }
 
-  glutSetWindowTitle(title);
-  glutSetIconTitle(title);
+  if (data_set) {
+    if (debug >= DEBUG_MAX) fprintf(stderr, "Displaying set %s.\n", data_set->id) ;
+    glutSetWindowTitle(data_set->id);
+    glutSetIconTitle(data_set->id);
+  }
 
-  if (isvisible) {
+  if (isvisible && data_set) {
     if (debug >= DEBUG_MAX) fputs("Displaying.\n", stderr) ;
-    display(data_set, num_items, num_values);    
+    display(data_set);    
     /*    glutPostRedisplay(); */
     if (debug >= DEBUG_MAX) fprintf(stderr, "Done (%d).\n", count) ;
   }
@@ -523,7 +412,7 @@ void do_update(void)
 void do_idle(void)
 {
 				/* Only update if there's new data. */
-  if (rotate || (!halted && DataReady(inputfp))) {	
+  if (rotate || (!halted && is_data_ready(inputfp))) {	
     if (debug >= DEBUG_MAX) fputs("Update display during idle time.\n", stderr);
     do_update();
   }
@@ -674,6 +563,7 @@ keyboard(unsigned char ch, int x, int y)
     break;
 
   default:
+    /*printf("key %d\n", ch); */
     break;
   }
 
@@ -764,7 +654,7 @@ static void
 drawBar(int color, float height)
 {
   int i, j;
-
+  /* printf("hieght=%f\n", height); */
   setColor(color);
 
   for (i = 0; i < 6; ++i) {
@@ -773,7 +663,8 @@ drawBar(int color, float height)
     for (j = 0; j < 4; j++) {
       if (cube_vertexes[i][j][1] >= 0) {
 	glVertex4f(cube_vertexes[i][j][0],
-		   height == 0 ? 0.00001 : height,
+		   height,
+		   //		   height == 0 ? 0.00001 : height,
 		   cube_vertexes[i][j][2],
 		   cube_vertexes[i][j][3]
 		   );
@@ -789,6 +680,9 @@ drawBar(int color, float height)
 drawIDname( char *id )
 {
   while (*(id) == ' ') ++id;	/* Skip spaces */
+  if (debug > DEBUG_MAX) {
+    fprintf(stderr, "draw %s\n", id);
+  }
   setColor(BLACK);
   glPushMatrix();
   glTranslatef(-0.3, -0.65, 0.7);
@@ -799,7 +693,7 @@ drawIDname( char *id )
 }
 
 void 
-drawOneItem(int num_bars, item_t *item)
+drawOneItem(Item *item)
 {
 
   int i;
@@ -808,16 +702,17 @@ drawOneItem(int num_bars, item_t *item)
 
   glPushMatrix();
 
-  glScalef(0.3, 1.0, 0.3);
+  glTranslatef(0.0, -0.66, 0.0); 
+  glScalef(0.3, 0.1, 0.3);
 
-  for (i = 0; i < num_bars; i++) {
+  for (i = 0; i <  item->dataCount; i++) {
     if (i >= MAX_CONTROLLABLE || !hide_value[i]) {
 				/* draw bar */	
       if (i < MAX_CONTROLLABLE && scale_value[i] != 1.0) {
-	drawBar(i + 1, item->value[i] * scale_value[i]);
+	drawBar(i + 1, item->datums[i]->value * scale_value[i]);
       }
       else {
-	drawBar(i + 1, item->value[i]);        /* draw bar */
+	drawBar(i + 1, item->datums[i]->value);        /* draw bar */
       }
     }
     glTranslatef(0.0, 0.0, -3.33);
@@ -849,14 +744,14 @@ drawGname( char *group, int length )
 }
 
 void 
-drawBase(char *gname, int length, int num_values)
+drawBase(char *gname, int length, Item *item)
 {
 
   int i, j;
-
+  int num_values = item->dataCount;
   drawGname(gname, length);
-  for (i = 0; i < num_values ; i++) {
-    drawValueName(value_name[i], i);
+  for (i = 0; i < num_values; i++) {
+    drawValueName(item->datums[i]->name, i);
   }
 
   glPushMatrix();
@@ -888,10 +783,11 @@ drawBase(char *gname, int length, int num_values)
 
 
 
-void showData (item_t *data_set, int num_items, int num_values)
+void showData (Dataset *data_set)
 {
+  int num_groups = data_set->groupCount;
+  int num_values = 0;
   int num_drawn;
-  int num_in_group;
   int step;
   char last_group[MAX_TEXT];
   float location;
@@ -900,31 +796,33 @@ void showData (item_t *data_set, int num_items, int num_values)
 
   glTranslatef(-(per_row / 2 + 1), 0.0, 0.0);
 
-  num_in_group = data_set[0].group_size;
-
   num_drawn = 0;
   step = 0;
 
-  while (num_drawn < num_items) {
+  while (num_drawn < num_groups) {
     
     int i;
     
-    num_in_group = data_set[num_drawn].group_size ;
-    drawBase(data_set[num_drawn].group,num_in_group,num_values);
+    int num_in_group = data_set->groups[num_drawn]->itemCount;
+    drawBase(data_set->groups[num_drawn]->id, num_in_group, data_set->groups[num_drawn]->items[0]);
     
     for (i = 0; i < num_in_group; i++) {
-      drawOneItem(num_values, &(data_set[num_drawn]));
-      num_drawn++;
+      int new_num_values = data_set->groups[num_drawn]->items[i]->dataCount;
+      if (new_num_values > num_values) {
+	num_values = new_num_values;
+      }
+      drawOneItem(data_set->groups[num_drawn]->items[i]);
       glTranslatef(1, 0.0, 0.0);
       step++;
     }
+    num_drawn++;
     
-    if (num_drawn < num_items) {
+    if (num_drawn < num_groups) {
 
-      if (step > per_row || step + data_set[num_drawn].group_size > per_row
+      if (step > per_row || step + data_set->groups[num_drawn]->itemCount > per_row
 	  && (!cram 
 	      || (step > (1 - cram_factor) * per_row)
-	      || (step + data_set[num_drawn].group_size > per_row * (1 + cram_factor)))
+	      || (step + data_set->groups[num_drawn]->itemCount > per_row * (1 + cram_factor)))
 	  ) {
 	glTranslatef(-step, 0.0, -(num_values + gap));
 	step = 0;
@@ -940,11 +838,9 @@ void showData (item_t *data_set, int num_items, int num_values)
 }
 
 void
-display(item_t *data_set, int num_items, int num_values)
+display(Dataset *data_set)
 {
 
-  static int first_time_thru = 1;
-
   projection(zoom_horz, zoom_vert);
 
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
@@ -962,7 +858,7 @@ display(item_t *data_set, int num_items, int num_values)
   }
 
   /* Draw bars */
-  showData(data_set, num_items, num_values);
+  showData(data_set);
 
   glDepthMask(GL_TRUE);
   if (useRGB) {
@@ -1154,6 +1050,8 @@ void usage(void)
   puts("gather | display [options]");
   puts(" ");
   puts("Options:");
+  puts("    -t proc|free|rup|<cmd> type of data to gather or command to exec.");
+  puts("    -host host rsh gather on host.");
   puts("    -witdh n   pixals.");
   puts("    -height n  pixals.");
   puts("    -perrow n  data sets per row.");
@@ -1189,6 +1087,7 @@ main(int argc, char **argv)
   int i;
 
   char hostname[MAX_TEXT] = "";
+  char *gather_type = GATHER_PROC;
 
   /* process commmand line args */
   for (i = 1; i < argc; ++i) {
@@ -1261,6 +1160,17 @@ main(int argc, char **argv)
     } else if (!strcmp("-v", argv[i])) {
       debug = DEBUG_MIN;
 
+    } else if (!strcmp("-t", argv[i])) {
+      gather_type = argv[++i];
+      if (!strcmp("proc", gather_type)) {
+	gather_type = GATHER_PROC;
+      }
+      else if (!strcmp("free", gather_type)) {
+	gather_type = GATHER_FREE;
+      }
+      else if (!strcmp("rup", gather_type)) {
+	gather_type = GATHER_RUP;
+      }
     } else {
       usage();
     }
@@ -1273,7 +1183,7 @@ main(int argc, char **argv)
     char *cmd = getenv(GR_GATHER_COMMAND);
     char command[MAX_LINE];
 
-    strcpy(command, cmd ? cmd : GATHER_COMMAND);
+    strcpy(command, cmd ? cmd : gather_type);
 
     if (sleep_seconds) {
       sprintf(command, "%s -sleep %d", command, sleep_seconds);
